/*
 * 
 * This is a remote cloner program for stationary HT12 code remotes.
 * It uses generic 433MHz transmitter and receiver modules plugged
 * directly into the headers at the given pins. You can change the
 * pins in the defined lines below.
 * 
 * To control the cloner, you can open the serial monitor to give commands.
 * Three commands are valid:
 * -Changing the code with a simple twelve-digit binary code:
 * (Example):'000000000000'
 * -Setting the cloner to continuously transmit the stored code:
 * (Example):'transmit'
 * -Setting the cloner to continuously receive and display the received code:
 * (Example):'receive'
 * 
 * With  the built-in EEPROM, the board will be able to remember the last
 * state of the device (store the code and mode). When starting up, it will
 * set the last active code to transmit and it will either be set to
 * receive or transmit based on the last state before power off.
 * 
 * For more info, visit the Instructables page about the project or read
 * the comments below in the program to understand its workings.
 * 
 * https://www.instructables.com/Garage-Remote-Cloner-and-Emulator/
 * 
 */

#include <EEPROM.h> // Using eeprom library to save codes.

// Wiring of the modules
// Transmitter
#define TVCC  A2  // 5V pin of the transmitter
#define TX    A1  // Data pin of the transmitter
#define TGND  A0  // GND pin of the transmitter

// Receiver
#define RVCC  5   // 5V pin of the receiver
#define RX    4   // Data pin of the receiver
#define RGND  2   // GND pin of the receiver
#define SilenceTime 200 // Low time after a signal is deemed valid

// Remote variables
long ZeroCounter = 0;   // Counts the consecutive zero states of the receiver
long SignalStart = 0;   // Records the starting time of the received signal
bool ReceivedCode[] = {0,0,0,0,0,0,0,0,0,0,0,0};    // Stores the received code to be displayed
bool CodeToTransmit[] = {0,0,0,0,0,0,0,0,0,0,0,0};  // Stores the code received from serial
bool Mode = 1;              // Mode selector: 0 means receive, 1 means transmit
#define PulseDuration 310   // Mefines the sub pulse length when transmitting.


void setup() {
  // Start serial communication
  Serial.begin(115200);
  Serial.println("ready!");
  
  // Setup transmitter
  pinMode(TVCC, OUTPUT);
  pinMode(TX, OUTPUT);
  pinMode(TGND, OUTPUT);
  // Power up transmitter
  digitalWrite(TGND, LOW);
  digitalWrite(TVCC, HIGH);
  
  // Setup receiver
  pinMode(RVCC, OUTPUT);
  pinMode(RX, INPUT);
  pinMode(RGND, OUTPUT);
  // Power up receiver
  digitalWrite(RGND, LOW);
  digitalWrite(RVCC, HIGH);

  // Read stored code and mode from eeprom and set it as active
  Serial.println("Previously stored code:");
  for(int i = 0; i < 12; i++){
    if(EEPROM.read(i) == 1){
      CodeToTransmit[i] = 1;
    }else{
      CodeToTransmit[i] = 0;
    }
    Serial.print(CodeToTransmit[i]);
    Serial.print(",");
  }
  
  if(EEPROM.read(12) == 1){
    Mode = 1;
  }else{
    Mode = 0;
  }
  Serial.println();
}

void Receiving(){
  while(digitalRead(RX)){}   // Wait for fist data bit
  SignalStart = micros() - SignalStart; // Calculate the length of the start pulse
  
  delayMicroseconds(480);
  for(int i = 0; i < 12; i++){     // Reading HT12 code (12 data bits)
    ReceivedCode[i] = digitalRead(RX);
    delayMicroseconds(930);
  }
  
  if((SignalStart > 250 && SignalStart < 350) == false){    // Throw away noise if start pulse is out of spec
    return;
  }else{                                                    // Display results if sync detected
    Serial.print("Pulse Length: " + (String)SignalStart + "us;   ");
    Serial.print("Code: ");
    for(int i = 0; i < 12; i++){
      Serial.print(ReceivedCode[i]);
    }
    Serial.println();
  }
}

void Transmitting(){
  /*
   * HT12 encoding: Each data bit is divided into three parts:
   * High, Low and Data. The first two sections are the same on each bit
   * and only the last section is changing based on the actual data bits.
   * Each section is 310 microseconds long, making one bit 930 microseconds.
   * After every 12 bits, there is a single stop pulse of 310 microseconds
   * and a silence period of 12 milliseconds. The patter repeates until
   * transmission is turned off.
   * 
   * As an example a '1' bit would be High-Low-High
   * And a '0' bit would be High-Low-Low
   */
  for(int i = 0; i < 12; i++){  // Transmitting 12 bits
    digitalWrite(TX, HIGH);
    delayMicroseconds(PulseDuration);
    digitalWrite(TX, LOW);
    delayMicroseconds(PulseDuration);
    digitalWrite(TX, CodeToTransmit[i]);
    delayMicroseconds(PulseDuration);
  }
    digitalWrite(TX, HIGH);             // Transmit stop bit
    delayMicroseconds(PulseDuration);
    digitalWrite(TX, LOW);
  digitalWrite(TX, LOW);
  delay(12);                            // Silence period of 12 milliseconds
}

void CommandHandler(String Command){    // The command handler deals with text based commands from the serial line
  if(Command == "transmit"){        // Set mode to 1 if transmit command received
    Serial.println("Transmitting");
    Mode = 1;
    EEPROM.write(12, 1);
  }else 
  if(Command == "receive"){       // Set mode to 0 if receive command is received
    Serial.println("Receiving");
    Mode = 0;
    EEPROM.write(12, 0);
  }else 
  if(Command.charAt(0) == '0' || Command.charAt(0) == '1'){   // Check if the input is a code and read it in
    for(int i = 0; i < 12; i++){
      if(Command.charAt(i) == '0'){   // Read code and save it to eeprom
        CodeToTransmit[i] = 0;
        EEPROM.write(i, 0);
      }else{
        CodeToTransmit[i] = 1;
        EEPROM.write(i, 1);
      }
    }
    Serial.print("Code set to: ");
    for(int i = 0; i < 12; i++){
      Serial.print(CodeToTransmit[i]);
    }
    Serial.println();
    Serial.println("Saved to eeprom for later.");
  }else{
    Serial.println("Bad Command");
  }
}

void loop() {
  if(Mode == 0){    // Deal with radio decoding if the mode is set to '0'
    if(digitalRead(RX) == LOW){          // If radio is LOW, note it down 
      ZeroCounter++;
    }else if(ZeroCounter > SilenceTime){    // If radio is HIGH, and is over set silence time, then check for a remote signal
      SignalStart = micros();          // Record time of start pulse
      Receiving();                     // Try to sync and decode incoming code
      ZeroCounter = 0;
    }
  }else{      // Continously transmit stored code while mode is set to '1'
    Transmitting();
  }
  
  if(Serial.available() > 0) {          // Check if serial data is available
    String Input = Serial.readString();   // Read serial until timeout
    Input.trim();
    Serial.println("Command: " + Input);
    CommandHandler(Input);    // Deal with commands after reading serial
  }
}
